#!/usr/bin/env python
import os

# Needed to make getaddrinfo() work in pymongo on Mac OS X
# Docs mention it's a better choice for Linux as well.
# This must be done asap in the worker
if "GEVENT_RESOLVER" not in os.environ:
    os.environ["GEVENT_RESOLVER"] = "ares"

# We must still monkey-patch the methods for job sub-pools.
from gevent import monkey
monkey.patch_all()

import sys
import argparse
import ujson as json
import datetime

sys.path.insert(0, os.getcwd())

from mrq import config, queue, utils
from mrq.context import set_current_config, set_current_job
from mrq.utils import load_class_by_path


def main():

    parser = argparse.ArgumentParser(description='Runs a task')

    cfg = config.get_config(parser=parser, config_type="run")
    cfg["is_cli"] = True
    set_current_config(cfg)

    if len(cfg["taskargs"]) == 1:
        params = json.loads(cfg["taskargs"][0])  # pylint: disable-msg=E1101
    else:
        params = {}

        # mrq-run taskpath a 1 b 2 => {"a": "1", "b": "2"}
        for group in utils.group_iter(cfg["taskargs"], n=2):
            if len(group) != 2:
                print "Number of arguments wasn't even"
                sys.exit(1)
            params[group[0]] = group[1]

    if cfg["async"]:
        ret = queue.send_task(
            cfg["taskpath"], params, sync=False, queue=cfg["queue"])
        print ret
    else:
        worker_class = load_class_by_path(cfg["worker_class"])
        job = worker_class.job_class(None)
        job.data = {
            "path": cfg["taskpath"],
            "params": params,
            "queue": cfg["queue"]
        }
        job.datestarted = datetime.datetime.utcnow()
        set_current_job(job)
        ret = job.perform()
        print json.dumps(ret)  # pylint: disable-msg=E1101

if __name__ == "__main__":
    main()




from collections import defaultdict
import datetime


class LogHandler(object):

    """ Job/Worker-aware log handler.

        We used the standard logging module before but it suffers from memory leaks
        when creating lots of logger objects.
    """

    def __init__(self, collection=None, quiet=False):
        self.reset()
        self.set_collection(collection)
        self.quiet = quiet

        # Import here to avoid import loop
        from .context import get_current_job
        self.get_current_job = get_current_job

    def get_logger(self, worker=None, job=None):
        return LoggerInterface(self, worker=worker, job=job)

    def set_collection(self, collection=None):
        self.collection = collection

    def reset(self):
        self.buffer = {
            "workers": defaultdict(list),
            "jobs": defaultdict(list)
        }

    def encode_if_unicode(self, string):
        if isinstance(string, unicode):
            return string.encode("utf-8", "replace")
        else:
            return string

    def decode_if_str(self, string):
        if isinstance(string, str):
            return string.decode("utf-8", "replace")
        else:
            return unicode(string)

    def log(self, level, *args, **kwargs):

        worker = kwargs.get("worker")
        job = kwargs.get("job")

        joined_unicode_args = u" ".join([self.decode_if_str(x) for x in args])
        formatted = u"%s [%s] %s" % (
            datetime.datetime.utcnow(), level.upper(), joined_unicode_args)

        if not self.quiet:
            try:
                print self.encode_if_unicode(formatted)
            except UnicodeDecodeError:
                print formatted

        if self.collection is False:
            return

        if worker is not None:
            self.buffer["workers"][worker].append(formatted)
        elif job is not None:
            if job == "current":
                job_object = self.get_current_job()
                if job_object:
                    self.buffer["jobs"][job_object.id].append(formatted)
            else:
                self.buffer["jobs"][job].append(formatted)

    def flush(self, w=0):

        # We may log some stuff before we are even connected to Mongo!
        if not self.collection:
            return

        inserts = [{
            "worker": k,
            "logs": "\n".join(v) + "\n"
        } for k, v in self.buffer["workers"].iteritems()] + [{
            "job": k,
            "logs": "\n".join(v) + "\n"
        } for k, v in self.buffer["jobs"].iteritems()]

        if len(inserts) == 0:
            return

        self.reset()

        try:
            self.collection.insert(inserts, w=w)
        except Exception as e:
            from mrq.context import get_current_worker
            self.log("debug", "Log insert failed: %s" % e, worker=get_current_worker())


class LoggerInterface(object):

    """ This object acts as a logger from python's logging module. """

    def __init__(self, handler, **kwargs):
        self._handler = handler
        self.kwargs = kwargs
        self.quiet = False

    @property
    def handler(self):
        if self._handler:
            return self._handler

        # Import here to avoid import loop
        from .context import get_current_worker
        worker = get_current_worker()

        if worker:
            return worker.log_handler
        else:
            self._handler = LogHandler(quiet=self.quiet)
            return self._handler

    def info(self, *args):
        self.handler.log("info", *args, **self.kwargs)

    def warning(self, *args):
        self.handler.log("warning", *args, **self.kwargs)

    def error(self, *args):
        self.handler.log("error", *args, **self.kwargs)

    def debug(self, *args):
        self.handler.log("debug", *args, **self.kwargs)



from gevent import monkey
monkey.patch_all()

from flask import Flask, request

import time
import os
import sys
from bson import ObjectId
import json
import argparse
from gevent.pywsgi import WSGIServer
from werkzeug.serving import run_with_reloader

sys.path.insert(0, os.getcwd())

from mrq.queue import send_task, Queue
from mrq.context import connections, set_current_config, get_current_config
from mrq.config import get_config

from mrq.dashboard.utils import jsonify, requires_auth

CURRENT_DIRECTORY = os.path.dirname(os.path.realpath(__file__))

parser = argparse.ArgumentParser(description='Start the MRQ dashboard')

cfg = get_config(parser=parser, config_type="dashboard")
set_current_config(cfg)

app = Flask("dashboard", static_folder=os.path.join(
    CURRENT_DIRECTORY, "static"))


@app.route('/')
@requires_auth
def root():
    return app.send_static_file("index.html")


@app.route('/api/datatables/taskexceptions')
@requires_auth
def api_task_exceptions():
    stats = list(connections.mongodb_jobs.mrq_jobs.aggregate([
        {"$match": {"status": "failed"}},
        {"$group": {"_id": {"path": "$path", "exceptiontype": "$exceptiontype"},
                    "jobs": {"$sum": 1}}},
    ])["result"])

    stats.sort(key=lambda x: -x["jobs"])
    start = int(request.args.get("iDisplayStart", 0))
    end = int(request.args.get("iDisplayLength", 20)) + start

    data = {
        "aaData": stats[start:end],
        "iTotalDisplayRecords": len(stats)
    }

    data["sEcho"] = request.args["sEcho"]

    return jsonify(data)


@app.route('/api/datatables/status')
@requires_auth
def api_jobstatuses():
    stats = list(connections.mongodb_jobs.mrq_jobs.aggregate([
        # https://jira.mongodb.org/browse/SERVER-11447
        {"$sort": {"status": 1}},
        {"$group": {"_id": "$status", "jobs": {"$sum": 1}}}
    ])["result"])

    stats.sort(key=lambda x: x["_id"])

    data = {
        "aaData": stats,
        "iTotalDisplayRecords": len(stats)
    }

    data["sEcho"] = request.args["sEcho"]

    return jsonify(data)


@app.route('/api/datatables/taskpaths')
@requires_auth
def api_taskpaths():
    stats = list(connections.mongodb_jobs.mrq_jobs.aggregate([
        {"$sort": {"path": 1}},  # https://jira.mongodb.org/browse/SERVER-11447
        {"$group": {"_id": "$path", "jobs": {"$sum": 1}}}
    ])["result"])

    stats.sort(key=lambda x: -x["jobs"])

    data = {
        "aaData": stats,
        "iTotalDisplayRecords": len(stats)
    }

    data["sEcho"] = request.args["sEcho"]

    return jsonify(data)


@app.route('/workers')
@requires_auth
def get_workers():
    collection = connections.mongodb_jobs.mrq_workers
    cursor = collection.find({"status": {"$ne": "stop"}})
    data = {"workers": list(cursor)}
    return jsonify(data)


def build_api_datatables_query(req):
    query = {}

    if req.args.get("redisqueue"):
        query["_id"] = {"$in": [ObjectId(x) for x in Queue(
            req.args.get("redisqueue")).list_job_ids(limit=1000)]}
    else:

        for param in ["queue", "path", "status", "exceptiontype"]:
            if req.args.get(param):
                query[param] = req.args.get(param)
        if req.args.get("id"):
            query["_id"] = ObjectId(req.args.get("id"))
        if req.args.get("worker"):
            query["worker"] = ObjectId(req.args.get("worker"))

        if req.args.get("params"):
            try:
                params_dict = json.loads(req.args.get("params"))

                for key in params_dict.keys():
                    query["params.%s" % key] = params_dict[key]
            except Exception as e:
                print "Error will converting form JSON: %s" % e

    return query


@app.route('/api/datatables/<unit>')
@requires_auth
def api_datatables(unit):

    # import time
    # time.sleep(5)

    collection = None
    sort = None
    skip = int(request.args.get("iDisplayStart", 0))
    limit = int(request.args.get("iDisplayLength", 20))

    if unit == "queues":

        queues = []
        for name, jobs in Queue.all().items():
            queue = Queue(name)
            q = {
                "name": name,
                "jobs": jobs,  # MongoDB size
                "size": queue.size(),  # Redis size
                "is_sorted": queue.is_sorted,
                "is_timed": queue.is_timed,
                "is_raw": queue.is_raw,
                "is_set": queue.is_set
            }

            if queue.is_sorted:
                q["graph_config"] = cfg.get("raw_queues", {}).get(name, {}).get("dashboard_graph", lambda: {
                    "start": time.time() - (7 * 24 * 3600),
                    "stop": time.time() + (7 * 24 * 3600),
                    "slices": 30
                } if queue.is_timed else {
                    "start": 0,
                    "stop": 100,
                    "slices": 30
                })()
                if q["graph_config"]:
                    q["graph"] = queue.get_sorted_graph(**q["graph_config"])

            if queue.is_timed:
                q["jobs_to_dequeue"] = queue.count_jobs_to_dequeue()

            queues.append(q)

        queues.sort(key=lambda x: -(x["jobs"] + x["size"]))

        data = {
            "aaData": queues,
            "iTotalDisplayRecords": len(queues)
        }

    elif unit == "workers":
        fields = None
        query = {"status": {"$nin": ["stop"]}}
        collection = connections.mongodb_jobs.mrq_workers
        sort = [("datestarted", -1)]

        if request.args.get("showstopped"):
            query = {}

    elif unit == "scheduled_jobs":
        collection = connections.mongodb_jobs.mrq_scheduled_jobs
        fields = None
        query = {}

    elif unit == "jobs":

        fields = None
        query = build_api_datatables_query(request)
        sort = [("_id", 1)]

        # We can't search easily params because we store it as decoded JSON in mongo :(
        # Add a string index?
        # if request.args.get("sSearch"):
        #   query.update(json.loads(request.args.get("sSearch")))
        collection = connections.mongodb_jobs.mrq_jobs

    if collection is not None:

        cursor = collection.find(query, fields=fields)

        if sort:
            cursor.sort(sort)

        if skip is not None:
            cursor.skip(skip)

        if limit is not None:
            cursor.limit(limit)

        data = {
            "aaData": list(cursor),
            "iTotalDisplayRecords": collection.find(query).count()
        }

    data["sEcho"] = request.args["sEcho"]

    return jsonify(data)


@app.route('/api/job/<job_id>/result')
@requires_auth
def api_job_result(job_id):
    collection = connections.mongodb_jobs.mrq_jobs

    job_data = collection.find_one(
        {"_id": ObjectId(job_id)}, fields=["result"])
    if not job_data:
        return jsonify({})

    return jsonify({
        "result": job_data.get("result")
    })


@app.route('/api/job/<job_id>/traceback')
@requires_auth
def api_job_traceback(job_id):
    collection = connections.mongodb_jobs.mrq_jobs
    job_data = collection.find_one(
        {"_id": ObjectId(job_id)}, fields=["traceback"])

    if not job_data:
        job_data = {}

    return jsonify({
        "traceback": job_data.get("traceback", "No exception raised")
    })


@app.route('/api/jobaction', methods=["POST"])
@requires_auth
def api_job_action():
    return jsonify({"job_id": send_task("mrq.basetasks.utils.JobAction",
                                        {k: v for k,
                                         v in request.form.iteritems()},
                                        queue=get_current_config()["dashboard_queue"])})


@app.route('/api/logs')
@requires_auth
def api_logs():
    collection = connections.mongodb_logs.mrq_logs

    if request.args.get("job"):
        query = {"job": ObjectId(request.args.get("job"))}
    elif request.args.get("worker"):
        query = {"worker": ObjectId(request.args.get("worker"))}
    else:
        raise Exception("No ID")

    if request.args.get("last_log_id"):
        query["_id"] = {"$gt": ObjectId(request.args.get("min_log_id"))}

    logs = list(collection.find(query, fields={"_id": 1, "logs": 1}))

    data = {
        "logs": "\n".join([lines["logs"] for lines in logs]),

        # Don't be surprised, this will send unexisting ObjectIds when we're up
        # to date!
        "last_log_id": logs[-1]["_id"] if len(logs) else ObjectId()
    }

    return jsonify(data)


def main():
    app.debug = True
    http = WSGIServer(('', int(os.environ.get("PORT", 5555))), app)
    run_with_reloader(http.serve_forever)


if __name__ == '__main__':
    main()




class Task(object):

    cancel_on_retry = False
    cancel_on_exception = False
    cancel_on_timeout = False

    # Are we the first task that a Job called?
    is_main_task = False

    def __init__(self):
        pass

    def run(self, params):
        raise NotImplementedError



from mrq.task import Task
from mrq.job import Job
from mrq.queue import Queue
from bson import ObjectId
from mrq.context import connections, get_current_config
from collections import defaultdict
from mrq.utils import group_iter
import datetime
import ujson as json


class JobAction(Task):

    def run(self, params):

        self.params = params
        self.collection = connections.mongodb_jobs.mrq_jobs

        query = self.build_query()

        return self.perform_action(self.params.get("action"), query)

    def build_query(self):
        query = {}
        if self.params.get("id"):
            query["_id"] = ObjectId(self.params.get("id"))

        # TODO use redis for queue
        for k in [
                "queue",
                "status",
                "worker",
                "path",
                "dateretry",
                "exceptiontype"]:
            if self.params.get(k):
                query[k] = self.params.get(k)

        if self.params.get("params"):
            params_dict = json.loads(self.params.get("params"))  # pylint: disable-msg=E1101

            for key in params_dict.keys():
                query["params.%s" % key] = params_dict[key]

        return query

    def perform_action(self, action, query):

        stats = {
            "requeued": 0,
            "cancelled": 0
        }

        if action == "cancel":

            # Finding the ttl here to expire is a bit hard because we may have mixed paths
            # and hence mixed ttls.
            # If we are cancelling by path, get this ttl
            if query.get("path"):

                task_def = get_current_config().get(
                    "tasks", {}).get(query["path"]) or {}
                result_ttl = task_def.get("result_ttl", Job.result_ttl)

            # If not, get the maxmimum ttl of all tasks.
            else:
                result_ttl = max(
                    [
                        Job.result_ttl] + [
                        task_def.get(
                            "result_ttl",
                            Job.result_ttl) for task_def in get_current_config().get(
                            "tasks",
                            {}).values()])

            now = datetime.datetime.utcnow()
            ret = self.collection.update(query, {"$set": {
                "status": "cancel",
                "dateexpires": now + datetime.timedelta(seconds=result_ttl),
                "dateupdated": now
            }}, multi=True)
            stats["cancelled"] = ret["n"]

            # Special case when emptying just by queue name: empty it directly!
            # In this case we could also loose some jobs that were queued after
            # the MongoDB update. They will be "lost" and requeued later like the other case
            # after the Redis BLPOP
            if query.keys() == ["queue"]:
                Queue(query["queue"]).empty()

        elif action == "requeue":

            # Requeue task by groups of maximum 1k items (if all in the same
            # queue)
            cursor = self.collection.find(query, fields=["_id", "queue"])

            # We must freeze the list because queries below would change it.
            # This could not fit in memory, research adding {"stats": {"$ne":
            # "queued"}} in the query
            fetched_jobs = list(cursor)

            for jobs in group_iter(fetched_jobs, n=1000):

                jobs_by_queue = defaultdict(list)
                for job in jobs:
                    jobs_by_queue[job["queue"]].append(job["_id"])
                    stats["requeued"] += 1

                for queue in jobs_by_queue:
                    self.collection.update({
                        "_id": {"$in": jobs_by_queue[queue]}
                    }, {"$set": {
                        "status": "queued",
                        "dateupdated": datetime.datetime.utcnow()
                    }}, multi=True)

                    # Between these two lines, jobs can become "lost" too.

                    Queue(queue).enqueue_job_ids(
                        [str(x) for x in jobs_by_queue[queue]])

        return stats



import gevent
import gevent.pool
import os
import signal
import datetime
import time
import socket
import traceback
import psutil
import sys
import ujson as json
from bson import ObjectId
from gevent.pywsgi import WSGIServer
from collections import defaultdict

from .job import Job
from .exceptions import JobTimeoutException, StopRequested, JobInterrupt
from .context import set_current_worker, set_current_job, get_current_job, connections, enable_greenlet_tracing
from .queue import Queue

# https://groups.google.com/forum/#!topic/gevent/EmZw9CVBC2g
# if "__pypy__" in sys.builtin_module_names:
#   def _reuse(self):
#       self._sock._reuse()
#   def _drop(self):
#       self._sock._drop()
#   gevent.socket.socket._reuse = _reuse
#   gevent.socket.socket._drop = _drop


class Worker(object):

    """ Main worker class. """

    # Allow easy overloading
    job_class = Job

    # Valid statuses:
    #       * init: General worker initialization
    #       * wait: Waiting for new jobs from Redis
    #       * spawn: Got some new jobs, greenlets are being spawned
    #       * full: All the worker pool is busy executing jobs
    #       * join: Waiting for current jobs to finish, no new one will be accepted
    #       * kill: Killing all current jobs
    #       * stop: Worker is stopped, no jobs should remain
    status = "init"

    mongodb_jobs = None
    mongodb_logs = None
    redis = None

    def __init__(self, config):

        self.config = config

        set_current_worker(self)

        if self.config.get("trace_greenlets"):
            enable_greenlet_tracing()

        self.datestarted = datetime.datetime.utcnow()
        self.queues = [x for x in self.config["queues"] if x]
        self.done_jobs = 0
        self.max_jobs = self.config["max_jobs"]

        self.connected = False  # MongoDB + Redis

        self.exitcode = 0
        self.process = psutil.Process(os.getpid())
        self.greenlet = gevent.getcurrent()

        self.id = ObjectId()
        if config["name"]:
            self.name = self.config["name"]
        else:
            self.name = self.make_name()

        self.pool_size = self.config["gevent"]

        from .logger import LogHandler
        self.log_handler = LogHandler(quiet=self.config["quiet"])
        self.log = self.log_handler.get_logger(worker=self.id)

        self.log.info(
            "Starting Gevent pool with %s worker greenlets (+ report, logs, adminhttp)" %
            self.pool_size)

        self.gevent_pool = gevent.pool.Pool(self.pool_size)

        # Keep references to main greenlets
        self.greenlets = {}

        # TODO by "tag"?
        self._traced_io = {
            "types": defaultdict(float),
            "tasks": defaultdict(float),
            "total": 0
        }

    def connect(self, force=False):

        if self.connected and not force:
            return

        # Accessing connections attributes will automatically connect
        self.redis = connections.redis
        self.mongodb_jobs = connections.mongodb_jobs

        if self.config["mongodb_logs"] == "0":
            self.log_handler.set_collection(False)  # Disable
        else:
            self.mongodb_logs = connections.mongodb_logs
            self.log_handler.set_collection(self.mongodb_logs.mrq_logs)

        self.connected = True

        # Be mindful that this is done each time a worker starts
        if not self.config["no_mongodb_ensure_indexes"]:
            self.ensure_indexes()

    def ensure_indexes(self):

        if self.mongodb_logs:

            self.mongodb_logs.mrq_logs.ensure_index(
                [("job", 1)], background=False)
            self.mongodb_logs.mrq_logs.ensure_index(
                [("worker", 1)], background=False, sparse=True)

        self.mongodb_jobs.mrq_workers.ensure_index(
            [("status", 1)], background=False)
        self.mongodb_jobs.mrq_workers.ensure_index(
            [("datereported", 1)], background=False, expireAfterSeconds=3600)

        self.mongodb_jobs.mrq_jobs.ensure_index(
            [("status", 1)], background=False)
        self.mongodb_jobs.mrq_jobs.ensure_index(
            [("path", 1), ("status", 1)], background=False)
        self.mongodb_jobs.mrq_jobs.ensure_index(
            [("worker", 1), ("status", 1)], background=False, sparse=True)
        self.mongodb_jobs.mrq_jobs.ensure_index(
            [("queue", 1), ("status", 1)], background=False)
        self.mongodb_jobs.mrq_jobs.ensure_index(
            [("dateexpires", 1)], sparse=True, background=False, expireAfterSeconds=7 * 24 * 3600)

        self.mongodb_jobs.mrq_scheduled_jobs.ensure_index(
            [("hash", 1)], unique=True, background=False, drop_dups=True)

        try:
            # This will be default in MongoDB 2.6
            self.mongodb_jobs.command(
                {"collMod": "mrq_jobs", "usePowerOf2Sizes": True})
            self.mongodb_jobs.command(
                {"collMod": "mrq_workers", "usePowerOf2Sizes": True})
        except:
            pass

    def make_name(self):
        """ Generate a human-readable name for this worker. """
        return "%s.%s" % (socket.gethostname().split(".")[0], os.getpid())

    def greenlet_scheduler(self):

        from .scheduler import Scheduler
        scheduler = Scheduler(self.mongodb_jobs.mrq_scheduled_jobs)

        scheduler.sync_tasks(self.config.get("scheduler_tasks") or [])

        while True:
            scheduler.check()
            time.sleep(int(self.config["scheduler_interval"]))

    def greenlet_report(self):
        """ This greenlet always runs in background to update current status in MongoDB every 10 seconds.

        Caution: it might get delayed when doing long blocking operations. Should we do this in a thread instead?
         """

        while True:
            try:
                self.report_worker()
            except Exception as e:
                self.log.error("When reporting: %s" % e)
            finally:
                time.sleep(self.config["report_interval"])

    def greenlet_logs(self):
        """ This greenlet always runs in background to update current logs in MongoDB every 10 seconds.

        Caution: it might get delayed when doing long blocking operations. Should we do this in a thread instead?
         """

        while True:
            try:
                self.flush_logs(w=0)
            except Exception as e:
                self.log.error("When flushing logs: %s" % e)
            finally:
                time.sleep(self.config["report_interval"])

    def get_memory(self):
        return self.process.get_memory_info().rss

    def get_worker_report(self):
        """ Returns a dict containing all the data we can about the current status of the worker and
            its jobs. """

        greenlets = []

        for greenlet in self.gevent_pool:
            g = {}
            short_stack = []
            stack = traceback.format_stack(greenlet.gr_frame)
            for s in stack[1:]:
                if "/gevent/hub.py" in s:
                    break
                short_stack.append(s)
            g["stack"] = short_stack

            job = get_current_job(id(greenlet))
            if job:
                job.save()
                if job.data:
                    g["path"] = job.data["path"]
                g["datestarted"] = job.datestarted
                g["id"] = str(job.id)
                g["time"] = getattr(greenlet, "_trace_time", 0)
                g["switches"] = getattr(greenlet, "_trace_switches", None)
                if job._current_io is not None:
                    g["io"] = job._current_io

            greenlets.append(g)

        # When faking network latency, all sockets are affected, including OS ones, but
        # we still want reliable reports so this is disabled.
        if self.config["add_network_latency"] != "0" and self.config["add_network_latency"]:
            cpu = {
                "user": 0,
                "system": 0,
                "percent": 0
            }
            mem_rss = 0
        else:
            cpu_times = self.process.get_cpu_times()
            cpu = {
                "user": cpu_times.user,
                "system": cpu_times.system,
                "percent": self.process.get_cpu_percent(0)
            }
            mem_rss = self.get_memory()

        # Avoid sharing passwords or sensitive config!
        whitelisted_config = [
            "max_jobs",
            "gevent",
            "processes",
            "queues",
            "scheduler",
            "name",
            "local_ip"
        ]

        io = None
        if self._traced_io:
            io = {}
            for k, v in self._traced_io.items():
                if k == "total":
                    io[k] = v
                else:
                    io[k] = sorted(v.items(), reverse=True, key=lambda x: x[1])

        return {
            "status": self.status,
            "config": {k: v for k, v in self.config.iteritems() if k in whitelisted_config},
            "done_jobs": self.done_jobs,
            "datestarted": self.datestarted,
            "datereported": datetime.datetime.utcnow(),
            "name": self.name,
            "io": io,
            "process": {
                "pid": self.process.pid,
                "cpu": cpu,
                "mem": {
                    "rss": mem_rss
                }
                # https://code.google.com/p/psutil/wiki/Documentation
                # get_open_files
                # get_connections
                # get_num_ctx_switches
                # get_num_fds
                # get_io_counters
                # get_nice
            },
            "jobs": greenlets
        }

    def report_worker(self, w=0):

        report = self.get_worker_report()

        if self.config["report_file"]:
            with open(self.config["report_file"], "wb") as f:
                f.write(json.dumps(report, ensure_ascii=False))  # pylint: disable-msg=E1101

        try:
            self.mongodb_jobs.mrq_workers.update({
                "_id": ObjectId(self.id)
            }, {"$set": report}, upsert=True, w=w)
        except Exception as e:
            self.log.debug("Worker report failed: %s" % e)

    def greenlet_admin(self):
        """ This greenlet is used to get status information about the worker when --admin_port was given
        """

        if self.config["processes"] > 1:
            self.log.debug(
                "Admin server disabled because of multiple processes.")
            return

        from flask import Flask
        from mrq.dashboard.utils import jsonify
        app = Flask("admin")

        @app.route('/')
        def route_index():
            report = self.get_worker_report()
            report.update({
                "_id": self.id
            })
            return jsonify(report)

        self.log.debug("Starting admin server on port %s" %
                       self.config["admin_port"])
        try:
            server = WSGIServer(
                ("0.0.0.0", self.config["admin_port"]), app, log=open(os.devnull, "w")
            )
            server.serve_forever()
        except Exception as e:
            self.log.debug("Error in admin server : %s" % e)

    def flush_logs(self, w=0):
        self.log_handler.flush(w=w)

    def work_loop(self):
        """Starts the work loop.

        """

        self.connect()

        self.status = "started"

        self.greenlets["report"] = gevent.spawn(self.greenlet_report)

        self.greenlets["logs"] = gevent.spawn(self.greenlet_logs)

        if self.config["scheduler"]:
            self.greenlets["scheduler"] = gevent.spawn(self.greenlet_scheduler)

        if self.config["admin_port"]:
            self.greenlets["admin"] = gevent.spawn(self.greenlet_admin)

        self.install_signal_handlers()

        # has_raw = any(q.is_raw or q.is_sorted for q in [Queue(x) for x in self.queues])

        try:

            wait_count = 0

            while True:

                while True:

                    free_pool_slots = self.gevent_pool.free_count()

                    if free_pool_slots > 0:
                        self.status = "wait"
                        break
                    self.status = "full"
                    gevent.sleep(0.01)

                quiet = not (wait_count % 20 == 0 or wait_count > 1000)

                if not quiet:
                    self.log.info('Fetching %s jobs from %s' % (free_pool_slots, self.queues))

                jobs = []

                for queue_name in self.queues:
                    queue = Queue(queue_name)

                    jobs += queue.dequeue_jobs(
                        max_jobs=free_pool_slots - len(jobs),
                        job_class=self.job_class,
                        worker=self
                    )

                    if len(jobs) >= free_pool_slots:
                        break

                for job in jobs:

                    # TODO investigate spawn_raw?
                    self.gevent_pool.spawn(self.perform_job, job)

                if self.max_jobs and self.max_jobs >= self.done_jobs:
                    self.log.info("Reached max_jobs=%s" % self.done_jobs)
                    break

                # We seem to have exhausted available jobs, we can sleep for a
                # while.
                if len(jobs) < free_pool_slots:
                    self.status = "wait"
                    wait_count += 1
                    gevent.sleep(min(1, 0.001 * wait_count))

        except StopRequested:
            pass

        finally:

            try:

                self.log.debug("Joining the greenlet pool...")
                self.status = "join"

                self.gevent_pool.join(timeout=None, raise_error=False)
                self.log.debug("Joined.")

            except StopRequested:
                pass

            self.status = "kill"

            self.gevent_pool.kill(exception=JobInterrupt, block=True)

            for g in self.greenlets:
                g_time = getattr(self.greenlets[g], "_trace_time", 0)
                g_switches = getattr(self.greenlets[g], "_trace_switches", None)
                self.greenlets[g].kill(block=True)
                self.log.debug(
                    "Greenlet for %s killed (%0.5fs, %s switches)." %
                    (g, g_time, g_switches))

            self.status = "stop"

            self.report_worker(w=1)
            self.flush_logs(w=1)

            g_time = getattr(self.greenlet, "_trace_time", 0)
            g_switches = getattr(self.greenlet, "_trace_switches", None)
            self.log.debug(
                "Exiting main worker greenlet (%0.5fs, %s switches)." %
                (g_time, g_switches))

        return self.exitcode

    def perform_job(self, job):
        """ Wraps a job.perform() call with timeout logic and exception handlers.

            This is the first call happening inside the greenlet.
        """

        if self.config["trace_memory"]:
            job.trace_memory_start()

        set_current_job(job)

        gevent_timeout = gevent.Timeout(
            job.timeout,
            JobTimeoutException(
                'Gevent Job exceeded maximum timeout value (%d seconds).' %
                job.timeout
            )
        )

        gevent_timeout.start()

        try:
            job.perform()

        except job.retry_on_exceptions:
            self.log.error("Caught exception => retry")
            job.save_retry(sys.exc_info()[1], exception=True)

        except job.cancel_on_exceptions:
            self.log.error("Job cancelled")
            job.save_status("cancel", exception=True)

        except JobTimeoutException:
            if job.task.cancel_on_timeout:
                self.log.error(
                    "Job timeouted after %s seconds, cancelled" % job.timeout)
                job.save_status("cancel", exception=True)
            else:
                self.log.error("Job timeouted after %s seconds" % job.timeout)
                job.save_status("timeout", exception=True)

        except JobInterrupt:
            job.save_status("interrupt", exception=True)

        except Exception:
            job.save_status("failed", exception=True)

        finally:
            gevent_timeout.cancel()
            set_current_job(None)

            self.done_jobs += 1

            if self.config["trace_memory"]:
                job.trace_memory_stop()

    def shutdown_graceful(self):
        """ Graceful shutdown: waits for all the jobs to finish. """

        # This is in the 'exitcodes' list in supervisord so processes
        # exiting gracefully won't be restarted.
        self.exitcode = 2

        self.log.info("Graceful shutdown...")
        raise StopRequested()

    def shutdown_now(self):
        """ Forced shutdown: interrupts all the jobs. """

        # This is in the 'exitcodes' list in supervisord so processes
        # exiting gracefully won't be restarted.
        self.exitcode = 3

        self.log.info("Forced shutdown...")
        self.status = "killing"

        self.gevent_pool.kill(exception=JobInterrupt, block=False)

        raise StopRequested()

    def install_signal_handlers(self):
        """ Handle events like Ctrl-C from the command line. """

        self.graceful_stop = False

        def request_shutdown_now():
            self.shutdown_now()

        def request_shutdown_graceful():

            # Second time CTRL-C, shutdown now
            if self.graceful_stop:
                request_shutdown_now()
            else:
                self.graceful_stop = True
                self.shutdown_graceful()

        # First time CTRL-C, try to shutdown gracefully
        gevent.signal(signal.SIGINT, request_shutdown_graceful)

        # User (or Heroku) requests a stop now, just mark tasks as interrupted.
        gevent.signal(signal.SIGTERM, request_shutdown_now)



from .context import get_current_job
import time
import random
import re


def patch_method(base_class, method_name, method):

    # Create a closure to store the old method
    def _patched_factory(old_method):
        def _mrq_patched_method(*args, **kwargs):
            return method(old_method, *args, **kwargs)
        return _mrq_patched_method

    old_method = getattr(base_class, method_name)
    setattr(base_class, method_name, _patched_factory(old_method))


def patch_io_all():
    """ Patch higher-level modules to provide insights on what is performing IO for each job. """

    patch_io_httplib()
    patch_io_redis()
    patch_io_pymongo_cursor()

    #patch_io_dns()
    #patch_io_subprocess()


def patch_pymongo(config):
    """ Monkey-patch pymongo's collections to add some logging """

    # Nothing to change!
    if not config["print_mongodb"] and not config["trace_io"]:
        return

    from termcolor import cprint

    # Print because we are very early and log() may not be ready yet.
    cprint("Monkey-patching MongoDB methods...", "white")

    def gen_monkey_patch(base_object, method):
        base_method = getattr(base_object, method)

        def mrq_monkey_patched(self, *args, **kwargs):
            if config["print_mongodb"]:
                if self.full_name in config.get(
                        "print_mongodb_hidden_collections",
                        []):
                    cprint("[MONGO] %s.%s%s %s" % (
                        self.full_name, method, "-hidden-", kwargs), "magenta")
                else:
                    cprint("[MONGO] %s.%s%s %s" %
                           (self.full_name, method, args, kwargs), "magenta")

            if config["trace_io"]:
                job = get_current_job()
                if job:
                    job.set_current_io({
                        "type": "mongodb.%s" % method,
                        "data": {
                            "collection": self.full_name
                        }
                        #"data": json.dumps(args)[0:300]  # Perf issue? All MongoDB data will get jsonified!
                    })

            try:
                ret = base_method(self, *args, **kwargs)
            finally:
                if config["trace_io"]:
                    job = get_current_job()
                    if job:
                        job.set_current_io(None)

            return ret

        # Needed to avoid breaking mongokit
        mrq_monkey_patched.__doc__ = method.__doc__

        return mrq_monkey_patched

    from pymongo.collection import Collection
    for method in ["find", "update", "insert", "remove", "find_and_modify"]:
        if getattr(Collection, method).__name__ != "mrq_monkey_patched":
            setattr(Collection, method, gen_monkey_patch(Collection, method))

    # MongoKit completely replaces the code from PyMongo's find() function, so we
    # need to monkey-patch that as well.
    try:
        from mongokit.collection import Collection as MongoKitCollection
        for method in ["find"]:
            if getattr(MongoKitCollection, method).__name__ != "mrq_monkey_patched":
                setattr(MongoKitCollection, method, gen_monkey_patch(MongoKitCollection, method))

    except ImportError:
        pass


# https://code.google.com/p/gevent/issues/detail?id=108
def patch_import():

    import types
    import gevent.coros
    import __builtin__

    orig_import = __builtin__.__import__
    import_lock = gevent.coros.RLock()

    def mrq_safe_import(*args, **kwargs):
        """
        Normally python protects imports against concurrency by doing some locking
        at the C level (at least, it does that in CPython).  This function just
        wraps the normal __import__ functionality in a recursive lock, ensuring that
        we're protected against greenlet import concurrency as well.
        """
        if len(args) > 0 and type(args[0]) not in [
                types.StringType,
                types.UnicodeType]:
            # if a builtin has been acquired as a bound instance method,
            # python knows not to pass 'self' when the method is called.
            # No such protection exists for monkey-patched builtins,
            # however, so this is necessary.
            args = args[1:]
        import_lock.acquire()
        try:
            result = orig_import(*args, **kwargs)
        finally:
            import_lock.release()
        return result

    builtins = __import__('__builtin__')
    if builtins.__import__.__name__ != "mrq_safe_import":
        builtins.__import__ = mrq_safe_import


def patch_network_latency(seconds=0.01):
    """ Add random latency to all I/O operations """

    # Accept float(0.1), "0.1", "0.1-0.2"
    def sleep():
        if isinstance(seconds, float):
            time.sleep(seconds)
        elif isinstance(seconds, basestring):
            if "-" in seconds:
                time.sleep(random.uniform(
                    float(seconds.split("-")[0]),
                    float(seconds.split("-")[1])
                ))
            else:
                time.sleep(float(seconds))

    def _patched_method(old_method, *args, **kwargs):
        sleep()
        return old_method(*args, **kwargs)

    socket_methods = [
        "send", "sendall", "sendto", "recv", "recvfrom", "recvfrom_into", "recv_into",
        "connect", "connect_ex", "close"
    ]

    from socket import socket as _socketmodule
    from gevent.socket import socket as _geventmodule
    from gevent.ssl import SSLSocket as _sslmodule

    for method in socket_methods:
        patch_method(_socketmodule, method, _patched_method)
        patch_method(_geventmodule, method, _patched_method)
        patch_method(_sslmodule, method, _patched_method)


def patch_io_redis():

    def execute_command(old_method, self, *args, **options):

        job = get_current_job()
        if job:
            job.set_current_io({
                "type": "redis.%s" % args[0].lower(),
                "data": {
                    "key": args[1] if len(args) > 1 else None
                }
            })

        try:
            ret = old_method(self, *args, **options)
        finally:
            if job:
                job.set_current_io(None)

        return ret

    from redis import StrictRedis

    patch_method(StrictRedis, "execute_command", execute_command)


def patch_io_httplib():
    """ Patch the base httplib.HTTPConnection class, which is used in most HTTP libraries
        like urllib2 or urllib3/requests. """

    def start(method, url):
        job = get_current_job()
        if job:
            job.set_current_io({
                "type": "http.%s" % method.lower(),
                "data": {
                    "url": url
                }
            })

    def stop():
        job = get_current_job()
        if job:
            job.set_current_io(None)

    class mrq_wrapped_socket(object):
        """ Socket-like object that keeps track of 'trace_args' and wraps our monitoring code
            around blocking I/O calls. """

        def __init__(self, obj, parent_connection):
            self._obj = obj
            self._parent_connection = parent_connection

            def _make_patched_method(method):
                def _patched_method(*args, **kwargs):

                    # In the case of HTTPS, we may connect() before having called conn.request()
                    # For requests/urllib3, we may need to plug ourselves at the connectionpool.urlopen level
                    if not hasattr(self._parent_connection, "_traced_args"):
                        return getattr(self._obj, method)(*args, **kwargs)

                    start(*self._parent_connection._traced_args)
                    try:
                        data = getattr(self._obj, method)(*args, **kwargs)
                    finally:
                        stop()
                    return data
                return _patched_method

            # Replace socket methods with instrumented ones
            for method in [

              # socket
              "send", "sendall", "sendto", "recv", "recvfrom", "recvfrom_into", "recv_into",
              "connect", "connect_ex", "close",

              # fileobject
              "read", "readline", "write", "writelines", "seek"
            ]:
                setattr(self, method, _make_patched_method(method))

        # Forward all other calls/attributes to the base socket
        def __getattr__(self, attr):
            # cprint(attr, "green")
            return getattr(self._obj, attr)

        def makefile(self, *args, **kwargs):
            newsock = self._obj.makefile(*args, **kwargs)
            return mrq_wrapped_socket(newsock, self._parent_connection)

    def request(old_method, self, method, url, body=None, headers={}):

        # This is for proxy support - TODO show that in dashboard?
        if re.search("^http(s?)\:\/\/", url):
            report_url = url
        else:
            protocol = "http"
            if hasattr(self, "key_file"):
                protocol = "https"

            report_url = "%s://%s%s%s" % (protocol, self.host, (":%s" % self.port) if self.port != 80 else "", url)

        self._traced_args = (method, report_url)
        res = old_method(self, method, url, body=body, headers=headers)
        return res

    def connect(old_method, self, *args, **kwargs):

        # In the case of HTTPS, we may connect() before having called conn.request()
        # For requests/urllib3, we may need to plug ourselves at the connectionpool.urlopen level
        if not hasattr(self, "_traced_args"):
            ret = old_method(self, *args, **kwargs)
        else:
            start(*self._traced_args)
            try:
                ret = old_method(self, *args, **kwargs)
            finally:
                stop()
        self.sock = mrq_wrapped_socket(self.sock, self)

        return ret

    from httplib import HTTPConnection, HTTPSConnection

    patch_method(HTTPConnection, "request", request)
    patch_method(HTTPConnection, "connect", connect)
    patch_method(HTTPSConnection, "connect", connect)

    # Try to patch requests & urllib3 as they are very popular python modules.
    try:
        from requests.packages.urllib3.connection import HTTPConnection, UnverifiedHTTPSConnection, VerifiedHTTPSConnection

        patch_method(HTTPConnection, "connect", connect)
        patch_method(UnverifiedHTTPSConnection, "connect", connect)
        patch_method(VerifiedHTTPSConnection, "connect", connect)

    except ImportError:
         pass

    try:
        from urllib3.connection import HTTPConnection, UnverifiedHTTPSConnection, VerifiedHTTPSConnection

        patch_method(HTTPConnection, "connect", connect)
        patch_method(UnverifiedHTTPSConnection, "connect", connect)
        patch_method(VerifiedHTTPSConnection, "connect", connect)

    except ImportError:
         pass


def patch_io_pymongo_cursor():

    from pymongo.cursor import Cursor

    class mrq_patched_pymongo_cursor(Cursor):

        # Some dark magic is needed here to cope with python's name mangling for private variables.
        def _Cursor__send_message(self, *args, **kwargs):
            # print self.__dict__
            job = get_current_job()

            if job:

                subtype = "find"
                collection = self._Cursor__collection.name  # pylint: disable-msg=E1101

                if collection == "$cmd":
                    items = self._Cursor__spec.items()  # pylint: disable-msg=E1101
                    if len(items) > 0:
                        subtype, collection = items[0]

                job.set_current_io({
                    "type": "mongodb.%s" % subtype,
                    "data": {
                        "collection": "%s.%s" % (self._Cursor__collection.database.name, collection)  # pylint: disable-msg=E1101
                    }
                })
            ret = Cursor._Cursor__send_message(self, *args, **kwargs)  # pylint: disable-msg=E1101

            if job:
                job.set_current_io(None)
            return ret

    import pymongo as pymongomodule
    if pymongomodule.cursor.Cursor.__name__ != "mrq_patched_pymongo_cursor":
        pymongomodule.cursor.Cursor = mrq_patched_pymongo_cursor

    if pymongomodule.collection.Cursor.__name__ != "mrq_patched_pymongo_cursor":
        pymongomodule.collection.Cursor = mrq_patched_pymongo_cursor
